pub mod merkle_stream;
pub mod tpm;
pub use merkle_stream::MerkleStream;
pub mod secret_stream;
pub use secret_stream::SecretStream;

use crate::{
    fmt, io, BigArray, BlockMeta, BlockPath, Deserialize, Display, Epoch, Formatter, Hashable,
    Principal, Principaled, Serialize, Writecap, WritecapBody,
};

use btserde::{self, from_vec, to_vec, write_to};
use foreign_types::ForeignType;
use openssl::{
    encrypt::{Decrypter as OsslDecrypter, Encrypter as OsslEncrypter},
    error::ErrorStack,
    hash::{hash, Hasher, MessageDigest},
    pkey::{HasPrivate, HasPublic, PKey},
    rand::rand_bytes,
    rsa::{Padding as OpensslPadding, Rsa as OsslRsa},
    sign::{Signer as OsslSigner, Verifier as OsslVerifier},
    symm::{decrypt as openssl_decrypt, encrypt as openssl_encrypt, Cipher, Crypter, Mode},
};
use serde::{
    de::{self, DeserializeOwned, Deserializer, SeqAccess, Visitor},
    ser::{SerializeStruct, Serializer},
};
use std::{marker::PhantomData, num::TryFromIntError, str::FromStr};
use strum_macros::{Display, EnumDiscriminants, EnumString};
use zeroize::ZeroizeOnDrop;

#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub struct Ciphertext<T> {
    data: Vec<u8>,
    phantom: PhantomData<T>,
}

impl<T> Ciphertext<T> {
    pub fn new(data: Vec<u8>) -> Ciphertext<T> {
        Ciphertext {
            data,
            phantom: PhantomData,
        }
    }
}

pub struct Signed<T> {
    _data: Vec<u8>,
    sig: Signature,
    phantom: PhantomData<T>,
}

impl<T> Signed<T> {
    pub fn new(data: Vec<u8>, sig: Signature) -> Signed<T> {
        Signed {
            _data: data,
            sig,
            phantom: PhantomData,
        }
    }
}

/// Errors that can occur during cryptographic operations.
#[derive(Debug)]
pub enum Error {
    NoReadCap,
    NoKeyAvailable,
    MissingPrivateKey,
    KeyVariantUnsupported,
    BlockNotEncrypted,
    InvalidHashFormat,
    InvalidSignature,
    IncorrectSize {
        expected: usize,
        actual: usize,
    },
    IndexOutOfBounds {
        index: usize,
        limit: usize,
    },
    IndivisibleSize {
        divisor: usize,
        actual: usize,
    },
    InvalidOffset {
        actual: usize,
        limit: usize,
    },
    HashCmpFailure,
    RootHashNotVerified,
    SignatureMismatch {
        actual: Principal,
        expected: Principal,
    },
    WritecapAuthzErr(WritecapAuthzErr),
    Serde(btserde::Error),
    Io(std::io::Error),
    Custom(Box<dyn std::fmt::Debug + Send + Sync>),
}

impl Error {
    pub(crate) fn custom<E: std::fmt::Debug + Send + Sync + 'static>(err: E) -> Self {
        Error::Custom(Box::new(err))
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        match self {
            Error::NoReadCap => write!(f, "no readcap"),
            Error::NoKeyAvailable => write!(f, "no key available"),
            Error::MissingPrivateKey => write!(f, "private key was missing"),
            Error::KeyVariantUnsupported => write!(f, "unsupported key variant"),
            Error::BlockNotEncrypted => write!(f, "block was not encrypted"),
            Error::InvalidHashFormat => write!(f, "invalid format"),
            Error::InvalidSignature => write!(f, "invalid signature"),
            Error::IncorrectSize { expected, actual } => {
                write!(f, "expected size {} but got {}", expected, actual)
            }
            Error::IndexOutOfBounds { index, limit } => write!(
                f,
                "index {} is out of bounds, it must be strictly less than {}",
                index, limit
            ),
            Error::IndivisibleSize { divisor, actual } => write!(
                f,
                "expected a size which is divisible by {} but got {}",
                divisor, actual
            ),
            Error::InvalidOffset { actual, limit } => write!(
                f,
                "offset {} is out of bounds, it must be strictly less than {}",
                actual, limit
            ),
            Error::HashCmpFailure => write!(f, "hash data are not equal"),
            Error::RootHashNotVerified => write!(f, "root hash is not verified"),
            Error::SignatureMismatch { actual, expected } => write!(
                f,
                "expected a signature from {expected} but found one from {actual}"
            ),
            Error::WritecapAuthzErr(err) => err.fmt(f),
            Error::Serde(err) => err.fmt(f),
            Error::Io(err) => err.fmt(f),
            Error::Custom(cus) => cus.fmt(f),
        }
    }
}

impl std::error::Error for Error {}

impl From<WritecapAuthzErr> for Error {
    fn from(err: WritecapAuthzErr) -> Self {
        Error::WritecapAuthzErr(err)
    }
}

impl From<ErrorStack> for Error {
    fn from(error: ErrorStack) -> Error {
        Error::custom(error)
    }
}

impl From<btserde::Error> for Error {
    fn from(error: btserde::Error) -> Error {
        Error::Serde(error)
    }
}

impl From<std::io::Error> for Error {
    fn from(error: std::io::Error) -> Error {
        Error::Io(error)
    }
}

impl From<TryFromIntError> for Error {
    fn from(err: TryFromIntError) -> Self {
        Error::custom(err)
    }
}

impl From<Error> for io::Error {
    fn from(err: Error) -> Self {
        io::Error::new(io::ErrorKind::Other, err)
    }
}

impl From<crate::Error> for Error {
    fn from(err: crate::Error) -> Self {
        Error::custom(err)
    }
}

pub(crate) type Result<T> = std::result::Result<T, Error>;

/// Returns an array of the given length filled with cryptographically strong random data.
pub fn rand_array<const LEN: usize>() -> Result<[u8; LEN]> {
    let mut array = [0; LEN];
    rand_bytes(&mut array)?;
    Ok(array)
}

/// Returns a vector of the given length with with cryptographically strong random data.
pub fn rand_vec(len: usize) -> Result<Vec<u8>> {
    let mut vec = vec![0; len];
    rand_bytes(&mut vec)?;
    Ok(vec)
}

/// A cryptographic hash.
#[derive(
    Debug,
    PartialEq,
    Eq,
    Serialize,
    Deserialize,
    Hashable,
    Clone,
    EnumDiscriminants,
    PartialOrd,
    Ord,
)]
#[strum_discriminants(derive(EnumString, Display, Serialize, Deserialize))]
#[strum_discriminants(name(HashKind))]
pub enum Hash {
    Sha2_256([u8; HashKind::Sha2_256.len()]),
    #[serde(with = "BigArray")]
    Sha2_512([u8; HashKind::Sha2_512.len()]),
}

impl Default for HashKind {
    fn default() -> HashKind {
        HashKind::Sha2_256
    }
}

impl Default for Hash {
    fn default() -> Self {
        HashKind::default().into()
    }
}

impl HashKind {
    #[allow(clippy::len_without_is_empty)]
    pub const fn len(self) -> usize {
        match self {
            HashKind::Sha2_256 => 32,
            HashKind::Sha2_512 => 64,
        }
    }

    pub fn digest<'a, I: Iterator<Item = &'a [u8]>>(self, dest: &mut [u8], parts: I) -> Result<()> {
        if dest.len() != self.len() {
            return Err(Error::IncorrectSize {
                expected: self.len(),
                actual: dest.len(),
            });
        }
        let mut hasher = Hasher::new(self.into())?;
        for part in parts {
            hasher.update(part)?;
        }
        let hash = hasher.finish()?;
        dest.copy_from_slice(&hash);
        Ok(())
    }
}

impl From<HashKind> for MessageDigest {
    fn from(kind: HashKind) -> Self {
        match kind {
            HashKind::Sha2_256 => MessageDigest::sha256(),
            HashKind::Sha2_512 => MessageDigest::sha512(),
        }
    }
}

impl Hash {
    /// The character that's used to separate a hash type from its value in its string
    /// representation.
    const HASH_SEP: char = ':';

    pub fn kind(&self) -> HashKind {
        HashKind::from(self)
    }
}

impl From<HashKind> for Hash {
    fn from(discriminant: HashKind) -> Hash {
        match discriminant {
            HashKind::Sha2_256 => Hash::Sha2_256([0u8; HashKind::Sha2_256.len()]),
            HashKind::Sha2_512 => Hash::Sha2_512([0u8; HashKind::Sha2_512.len()]),
        }
    }
}

impl AsRef<[u8]> for Hash {
    fn as_ref(&self) -> &[u8] {
        match self {
            Hash::Sha2_256(arr) => arr,
            Hash::Sha2_512(arr) => arr,
        }
    }
}

impl AsMut<[u8]> for Hash {
    fn as_mut(&mut self) -> &mut [u8] {
        match self {
            Hash::Sha2_256(arr) => arr,
            Hash::Sha2_512(arr) => arr,
        }
    }
}

impl TryFrom<&str> for Hash {
    type Error = Error;
    fn try_from(string: &str) -> Result<Hash> {
        let mut split: Vec<&str> = string.split(Self::HASH_SEP).collect();
        if split.len() != 2 {
            return Err(Error::InvalidHashFormat);
        };
        let second = split.pop().ok_or(Error::InvalidHashFormat)?;
        let first = split.pop().ok_or(Error::InvalidHashFormat)?;
        let mut hash = Hash::from(HashKind::from_str(first).map_err(|_| Error::InvalidHashFormat)?);
        base64_url::decode_to_slice(second, hash.as_mut()).map_err(|_| Error::InvalidHashFormat)?;
        Ok(hash)
    }
}

impl Display for Hash {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let hash_type: HashKind = self.into();
        let hash_data = base64_url::encode(self.as_ref());
        write!(f, "{}{}{}", hash_type, Hash::HASH_SEP, hash_data)
    }
}

/// A cryptographic signature.
#[derive(Debug, PartialEq, Serialize, Deserialize, Clone, Default)]
pub struct Signature {
    kind: Sign,
    data: Vec<u8>,
}

impl Signature {
    pub fn empty(kind: Sign) -> Signature {
        let data = vec![0; kind.key_len() as usize];
        Signature { kind, data }
    }

    #[cfg(test)]
    pub(crate) fn copy_from(kind: Sign, from: &[u8]) -> Signature {
        let mut data = vec![0; kind.key_len() as usize];
        data.as_mut_slice().copy_from_slice(from);
        Signature { kind, data }
    }

    pub fn as_slice(&self) -> &[u8] {
        self.data.as_slice()
    }

    pub fn as_mut_slice(&mut self) -> &mut [u8] {
        self.data.as_mut_slice()
    }
}

impl AsRef<[u8]> for Signature {
    fn as_ref(&self) -> &[u8] {
        self.as_slice()
    }
}

impl AsMut<[u8]> for Signature {
    fn as_mut(&mut self) -> &mut [u8] {
        self.as_mut_slice()
    }
}

#[derive(Serialize, Deserialize)]
struct TaggedCiphertext<T, U> {
    aad: U,
    ciphertext: Ciphertext<T>,
    tag: Vec<u8>,
}

#[derive(EnumDiscriminants, ZeroizeOnDrop)]
#[strum_discriminants(name(AeadKeyKind))]
#[strum_discriminants(derive(Serialize, Deserialize))]
pub enum AeadKey {
    AesGcm256 {
        key: [u8; AeadKeyKind::AesGcm256.key_len()],
        iv: [u8; AeadKeyKind::AesGcm256.iv_len()],
    },
}

impl AeadKeyKind {
    const fn key_len(self) -> usize {
        match self {
            AeadKeyKind::AesGcm256 => 32,
        }
    }

    const fn iv_len(self) -> usize {
        match self {
            AeadKeyKind::AesGcm256 => 16,
        }
    }
}

fn array_from<const N: usize>(slice: &[u8]) -> Result<[u8; N]> {
    let slice_len = slice.len();
    if N != slice_len {
        return Err(Error::IncorrectSize {
            actual: slice_len,
            expected: N,
        });
    }
    let mut array = [0u8; N];
    array.copy_from_slice(slice);
    Ok(array)
}

impl AeadKey {
    pub fn new(kind: AeadKeyKind) -> Result<AeadKey> {
        match kind {
            AeadKeyKind::AesGcm256 => Ok(AeadKey::AesGcm256 {
                key: rand_array()?,
                iv: rand_array()?,
            }),
        }
    }

    fn copy_components(kind: AeadKeyKind, key_buf: &[u8], iv_buf: &[u8]) -> Result<AeadKey> {
        match kind {
            AeadKeyKind::AesGcm256 => Ok(AeadKey::AesGcm256 {
                key: array_from(key_buf)?,
                iv: array_from(iv_buf)?,
            }),
        }
    }

    fn encrypt<T: Serialize + DeserializeOwned, U: Serialize + DeserializeOwned>(
        &self,
        aad: U,
        plaintext: &T,
    ) -> Result<TaggedCiphertext<T, U>> {
        let (cipher, key, iv, mut tag) = match self {
            AeadKey::AesGcm256 { key, iv } => (
                Cipher::aes_256_gcm(),
                key.as_slice(),
                iv.as_slice(),
                vec![0u8; 16],
            ),
        };

        let aad_data = to_vec(&aad)?;
        let plaintext_buf = to_vec(&plaintext)?;
        let mut ciphertext = vec![0u8; plaintext_buf.len() + cipher.block_size()];
        let mut crypter = Crypter::new(cipher, Mode::Encrypt, key, Some(iv))?;
        crypter.aad_update(&aad_data)?;
        let mut count = crypter.update(&plaintext_buf, &mut ciphertext)?;
        count += crypter.finalize(&mut ciphertext[count..])?;
        ciphertext.truncate(count);
        crypter.get_tag(&mut tag)?;

        Ok(TaggedCiphertext {
            aad,
            ciphertext: Ciphertext::new(ciphertext),
            tag,
        })
    }

    fn decrypt<T: Serialize + DeserializeOwned, U: Serialize + DeserializeOwned>(
        &self,
        tagged: &TaggedCiphertext<T, U>,
    ) -> Result<T> {
        let ciphertext = &tagged.ciphertext.data;
        let (cipher, key, iv) = match self {
            AeadKey::AesGcm256 { key, iv } => {
                (Cipher::aes_256_gcm(), key.as_slice(), iv.as_slice())
            }
        };

        let mut plaintext = vec![0u8; ciphertext.len() + cipher.block_size()];
        let mut crypter = Crypter::new(cipher, Mode::Decrypt, key, Some(iv))?;
        crypter.set_tag(&tagged.tag)?;
        let aad_buf = to_vec(&tagged.aad)?;
        crypter.aad_update(&aad_buf)?;
        let mut count = crypter.update(ciphertext, &mut plaintext)?;
        count += crypter.finalize(&mut plaintext[count..])?;
        plaintext.truncate(count);

        Ok(from_vec(&plaintext)?)
    }
}

#[derive(Debug, PartialEq, Serialize, Deserialize, Clone, EnumDiscriminants, ZeroizeOnDrop)]
#[strum_discriminants(name(SymKeyKind))]
pub enum SymKey {
    /// A key for the AES 256 cipher in Cipher Block Chaining mode. Note that this includes the
    /// initialization vector, so that a value of this variant contains all the information needed
    /// to fully initialize a cipher context.
    Aes256Cbc { key: [u8; 32], iv: [u8; 16] },
    /// A key for the AES 256 cipher in counter mode.
    Aes256Ctr { key: [u8; 32], iv: [u8; 16] },
}

struct SymParams<'a> {
    cipher: Cipher,
    key: &'a [u8],
    iv: Option<&'a [u8]>,
}

impl SymKey {
    pub(crate) fn generate(kind: SymKeyKind) -> Result<SymKey> {
        match kind {
            SymKeyKind::Aes256Cbc => Ok(SymKey::Aes256Cbc {
                key: rand_array()?,
                iv: rand_array()?,
            }),
            SymKeyKind::Aes256Ctr => Ok(SymKey::Aes256Ctr {
                key: rand_array()?,
                iv: rand_array()?,
            }),
        }
    }

    fn params(&self) -> SymParams {
        let (cipher, key, iv) = match self {
            SymKey::Aes256Cbc { key, iv } => (Cipher::aes_256_cbc(), key, Some(iv.as_slice())),
            SymKey::Aes256Ctr { key, iv } => (Cipher::aes_256_ctr(), key, Some(iv.as_slice())),
        };
        SymParams { cipher, key, iv }
    }

    fn block_size(&self) -> usize {
        let SymParams { cipher, .. } = self.params();
        cipher.block_size()
    }

    // The number of bytes that the plaintext expands by when encrypted.
    fn expansion_sz(&self) -> usize {
        match self {
            SymKey::Aes256Cbc { .. } => 16,
            SymKey::Aes256Ctr { .. } => 0,
        }
    }

    fn to_encrypter(&self) -> Result<Crypter> {
        let SymParams { cipher, key, iv } = self.params();
        Ok(Crypter::new(cipher, Mode::Encrypt, key, iv)?)
    }

    fn to_decrypter(&self) -> Result<Crypter> {
        let SymParams { cipher, key, iv } = self.params();
        Ok(Crypter::new(cipher, Mode::Decrypt, key, iv)?)
    }

    pub fn key_slice(&self) -> &[u8] {
        let SymParams { key, .. } = self.params();
        key
    }

    pub fn iv_slice(&self) -> Option<&[u8]> {
        let SymParams { iv, .. } = self.params();
        iv
    }
}

impl Encrypter for SymKey {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        let SymParams { cipher, key, iv } = self.params();
        Ok(openssl_encrypt(cipher, key, iv, slice)?)
    }
}

impl Decrypter for SymKey {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        let SymParams { cipher, key, iv } = self.params();
        Ok(openssl_decrypt(cipher, key, iv, slice)?)
    }
}

impl Default for SymKeyKind {
    fn default() -> Self {
        SymKeyKind::Aes256Ctr
    }
}

#[repr(u32)]
#[derive(Debug, Display, Clone, Copy, Serialize, Deserialize, PartialEq, Eq)]
pub enum KeyLen {
    Bits256 = 32,
    Bits512 = 64,
    Bits2048 = 256,
    Bits3072 = 384,
    Bits4096 = 512,
}

impl KeyLen {
    const fn bits(self) -> u32 {
        8 * self as u32
    }
}

pub trait Scheme:
    for<'de> Deserialize<'de> + Serialize + Copy + std::fmt::Debug + PartialEq + Into<Self::Kind>
{
    type Kind: Scheme;
    fn as_enum(self) -> SchemeKind;
    fn message_digest(&self) -> MessageDigest;
    fn padding(&self) -> Option<OpensslPadding>;
    fn public_from_der(self, der: &[u8]) -> Result<PKey<Public>>;
    fn private_from_der(self, der: &[u8]) -> Result<PKey<Private>>;
    fn generate(self) -> Result<AsymKeyPair<Self::Kind>>;
    fn key_len(self) -> KeyLen;
}

pub enum SchemeKind {
    Sign(Sign),
    Encrypt(Encrypt),
}

#[derive(Deserialize, Serialize, Clone, Debug, PartialEq, Copy)]
pub enum Encrypt {
    RsaEsOaep(RsaEsOaep),
}

impl Scheme for Encrypt {
    type Kind = Encrypt;

    fn as_enum(self) -> SchemeKind {
        SchemeKind::Encrypt(self)
    }

    fn message_digest(&self) -> MessageDigest {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.message_digest(),
        }
    }

    fn padding(&self) -> Option<OpensslPadding> {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.padding(),
        }
    }

    fn public_from_der(self, der: &[u8]) -> Result<PKey<Public>> {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.public_from_der(der),
        }
    }

    fn private_from_der(self, der: &[u8]) -> Result<PKey<Private>> {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.private_from_der(der),
        }
    }

    fn generate(self) -> Result<AsymKeyPair<Self::Kind>> {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.generate(),
        }
    }

    fn key_len(self) -> KeyLen {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.key_len(),
        }
    }
}

impl Encrypt {
    pub const RSA_OAEP_2048_SHA_256: Encrypt = Encrypt::RsaEsOaep(RsaEsOaep {
        key_len: KeyLen::Bits2048,
        hash_kind: HashKind::Sha2_256,
    });

    pub const RSA_OAEP_3072_SHA_256: Encrypt = Encrypt::RsaEsOaep(RsaEsOaep {
        key_len: KeyLen::Bits3072,
        hash_kind: HashKind::Sha2_256,
    });
}

#[derive(Deserialize, Serialize, Clone, Debug, PartialEq, Copy)]
pub enum Sign {
    RsaSsaPss(RsaSsaPss),
}

impl Default for Sign {
    fn default() -> Self {
        Self::RSA_PSS_2048_SHA_256
    }
}

impl Scheme for Sign {
    type Kind = Sign;

    fn as_enum(self) -> SchemeKind {
        SchemeKind::Sign(self)
    }

    fn message_digest(&self) -> MessageDigest {
        match self {
            Sign::RsaSsaPss(inner) => inner.message_digest(),
        }
    }

    fn padding(&self) -> Option<OpensslPadding> {
        match self {
            Sign::RsaSsaPss(inner) => inner.padding(),
        }
    }

    fn public_from_der(self, der: &[u8]) -> Result<PKey<Public>> {
        match self {
            Sign::RsaSsaPss(inner) => inner.public_from_der(der),
        }
    }

    fn private_from_der(self, der: &[u8]) -> Result<PKey<Private>> {
        match self {
            Sign::RsaSsaPss(inner) => inner.private_from_der(der),
        }
    }

    fn generate(self) -> Result<AsymKeyPair<Self::Kind>> {
        match self {
            Sign::RsaSsaPss(inner) => inner.generate(),
        }
    }

    fn key_len(self) -> KeyLen {
        self.key_len_const()
    }
}

impl Sign {
    pub const RSA_PSS_2048_SHA_256: Sign = Sign::RsaSsaPss(RsaSsaPss {
        key_bits: KeyLen::Bits2048,
        hash_kind: HashKind::Sha2_256,
    });

    pub const RSA_PSS_3072_SHA_256: Sign = Sign::RsaSsaPss(RsaSsaPss {
        key_bits: KeyLen::Bits3072,
        hash_kind: HashKind::Sha2_256,
    });

    const fn key_len_const(self) -> KeyLen {
        match self {
            Sign::RsaSsaPss(inner) => inner.key_bits,
        }
    }
}

enum Rsa {}

impl Rsa {
    /// The default public exponent to use for generated RSA keys.
    const EXP: u32 = 65537; // 2**16 + 1

    fn generate<S: Scheme>(scheme: S) -> Result<AsymKeyPair<S>> {
        let key = OsslRsa::generate(scheme.key_len().bits())?;
        // TODO: Separating the keys this way seems inefficient. Investigate alternatives.
        let public_der = key.public_key_to_der()?;
        let private_der = key.private_key_to_der()?;
        let public = AsymKey::<Public, S>::new(scheme, &public_der)?;
        let private = AsymKey::<Private, S>::new(scheme, &private_der)?;
        Ok(AsymKeyPair { public, private })
    }
}

#[derive(Deserialize, Serialize, Clone, Debug, PartialEq, Copy)]
pub struct RsaEsOaep {
    key_len: KeyLen,
    hash_kind: HashKind,
}

impl Scheme for RsaEsOaep {
    type Kind = Encrypt;

    fn as_enum(self) -> SchemeKind {
        SchemeKind::Encrypt(self.into())
    }

    fn message_digest(&self) -> MessageDigest {
        self.hash_kind.into()
    }

    fn padding(&self) -> Option<OpensslPadding> {
        Some(OpensslPadding::PKCS1_OAEP)
    }

    fn public_from_der(self, der: &[u8]) -> Result<PKey<Public>> {
        Ok(PKey::public_key_from_der(der)?.conv_pub())
    }

    fn private_from_der(self, der: &[u8]) -> Result<PKey<Private>> {
        Ok(PKey::private_key_from_der(der)?.conv_priv())
    }

    fn generate(self) -> Result<AsymKeyPair<Self::Kind>> {
        Rsa::generate(self.into())
    }

    fn key_len(self) -> KeyLen {
        self.key_len
    }
}

impl From<RsaEsOaep> for Encrypt {
    fn from(scheme: RsaEsOaep) -> Self {
        Encrypt::RsaEsOaep(scheme)
    }
}

#[derive(Deserialize, Serialize, Clone, Debug, PartialEq, Copy)]
pub struct RsaSsaPss {
    key_bits: KeyLen,
    hash_kind: HashKind,
}

impl Scheme for RsaSsaPss {
    type Kind = Sign;

    fn as_enum(self) -> SchemeKind {
        SchemeKind::Sign(self.into())
    }

    fn message_digest(&self) -> MessageDigest {
        self.hash_kind.into()
    }

    fn padding(&self) -> Option<OpensslPadding> {
        Some(OpensslPadding::PKCS1_PSS)
    }

    fn public_from_der(self, der: &[u8]) -> Result<PKey<Public>> {
        Ok(PKey::public_key_from_der(der)?.conv_pub())
    }

    fn private_from_der(self, der: &[u8]) -> Result<PKey<Private>> {
        Ok(PKey::private_key_from_der(der)?.conv_priv())
    }

    fn generate(self) -> Result<AsymKeyPair<Self::Kind>> {
        Rsa::generate(self.into())
    }

    fn key_len(self) -> KeyLen {
        self.key_bits
    }
}

impl From<RsaSsaPss> for Sign {
    fn from(scheme: RsaSsaPss) -> Self {
        Sign::RsaSsaPss(scheme)
    }
}

/// Marker trait for the `Public` and `Private` key privacy types.
pub trait KeyPrivacy {}

/// Represents keys which can be shared freely.
#[derive(Clone, Debug)]
pub enum Public {}

impl KeyPrivacy for Public {}

unsafe impl HasPublic for Public {}

#[derive(Debug, Clone)]
/// Represents keys which must be kept confidential.
pub enum Private {}

impl KeyPrivacy for Private {}

unsafe impl HasPrivate for Private {}

trait PKeyExt<T> {
    /// Converts a PKey<T> to a PKey<U>. This hack allows for converting between openssl's
    /// Public and Private types and ours.
    fn conv_pkey<U>(self) -> PKey<U>;

    /// Convert from openssl's Public type to `crypto::Public`.
    fn conv_pub(self) -> PKey<Public>;

    /// Convert from openssl's Private type to `crypto::Private`.
    fn conv_priv(self) -> PKey<Private>;
}

impl<T> PKeyExt<T> for PKey<T> {
    fn conv_pkey<U>(self) -> PKey<U> {
        let ptr = self.as_ptr();
        let new_pkey = unsafe { PKey::from_ptr(ptr) };
        std::mem::forget(self);
        new_pkey
    }

    fn conv_pub(self) -> PKey<Public> {
        self.conv_pkey()
    }

    fn conv_priv(self) -> PKey<Private> {
        self.conv_pkey()
    }
}

/// Represents any kind of asymmetric key.
#[derive(Debug, Clone)]
pub struct AsymKey<P, S> {
    scheme: S,
    pkey: PKey<P>,
}

impl<P, S: Copy> AsymKey<P, S> {
    pub fn scheme(&self) -> S {
        self.scheme
    }
}

pub type AsymKeyPub<S> = AsymKey<Public, S>;

impl<S: Scheme> AsymKey<Public, S> {
    pub(crate) fn new(scheme: S, der: &[u8]) -> Result<AsymKey<Public, S>> {
        let pkey = scheme.public_from_der(der)?;
        Ok(AsymKey { scheme, pkey })
    }
}

impl<S: Scheme> AsymKey<Private, S> {
    pub(crate) fn new(scheme: S, der: &[u8]) -> Result<AsymKey<Private, S>> {
        let pkey = scheme.private_from_der(der)?;
        Ok(AsymKey { scheme, pkey })
    }
}

impl<'de, S: Scheme> Deserialize<'de> for AsymKey<Public, S> {
    fn deserialize<D: Deserializer<'de>>(d: D) -> std::result::Result<Self, D::Error> {
        const FIELDS: &[&str] = &["scheme", "pkey"];

        struct StructVisitor<S: Scheme>(PhantomData<S>);

        impl<'de, S: Scheme> Visitor<'de> for StructVisitor<S> {
            type Value = AsymKey<Public, S>;

            fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
                formatter.write_fmt(format_args!("struct {}", stringify!(AsymKey)))
            }

            fn visit_seq<V: SeqAccess<'de>>(
                self,
                mut seq: V,
            ) -> std::result::Result<Self::Value, V::Error> {
                let scheme: S = seq
                    .next_element()?
                    .ok_or_else(|| de::Error::missing_field(FIELDS[0]))?;
                let der: Vec<u8> = seq
                    .next_element()?
                    .ok_or_else(|| de::Error::missing_field(FIELDS[1]))?;
                AsymKey::<Public, _>::new(scheme, der.as_slice()).map_err(de::Error::custom)
            }
        }

        d.deserialize_struct(stringify!(AsymKey), FIELDS, StructVisitor(PhantomData))
    }
}

impl<S: Scheme> Serialize for AsymKey<Public, S> {
    fn serialize<T: Serializer>(&self, s: T) -> std::result::Result<T::Ok, T::Error> {
        let mut struct_s = s.serialize_struct(stringify!(AsymKey), 2)?;
        struct_s.serialize_field("scheme", &self.scheme)?;
        let der = self.pkey.public_key_to_der().unwrap();
        struct_s.serialize_field("pkey", der.as_slice())?;
        struct_s.end()
    }
}

impl<S: Scheme> PartialEq for AsymKey<Public, S> {
    fn eq(&self, other: &Self) -> bool {
        self.scheme == other.scheme && self.pkey.public_eq(&other.pkey)
    }
}

impl Principaled for AsymKey<Public, Sign> {
    fn principal_of_kind(&self, kind: HashKind) -> Principal {
        let der = self.pkey.public_key_to_der().unwrap();
        let bytes = hash(kind.into(), der.as_slice()).unwrap();
        let mut hash_buf = Hash::from(kind);
        hash_buf.as_mut().copy_from_slice(&bytes);
        Principal(hash_buf)
    }
}

impl Encrypter for AsymKey<Public, Encrypt> {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        let mut encrypter = OsslEncrypter::new(&self.pkey)?;
        if let Some(padding) = self.scheme.padding() {
            encrypter.set_rsa_padding(padding)?;
        }
        {
            let Encrypt::RsaEsOaep(inner) = self.scheme;
            encrypter.set_rsa_oaep_md(inner.message_digest())?;
        }
        let buffer_len = encrypter.encrypt_len(slice)?;
        let mut ciphertext = vec![0; buffer_len];
        let ciphertext_len = encrypter.encrypt(slice, &mut ciphertext)?;
        ciphertext.truncate(ciphertext_len);
        Ok(ciphertext)
    }
}

impl Decrypter for AsymKey<Private, Encrypt> {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        let mut decrypter = OsslDecrypter::new(&self.pkey)?;
        if let Some(padding) = self.scheme.padding() {
            decrypter.set_rsa_padding(padding)?;
        }
        {
            let Encrypt::RsaEsOaep(inner) = self.scheme;
            decrypter.set_rsa_oaep_md(inner.message_digest())?;
        }
        let buffer_len = decrypter.decrypt_len(slice)?;
        let mut plaintext = vec![0; buffer_len];
        let plaintext_len = decrypter.decrypt(slice, &mut plaintext)?;
        plaintext.truncate(plaintext_len);
        Ok(plaintext)
    }
}

impl Signer for AsymKey<Private, Sign> {
    fn sign<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I) -> Result<Signature> {
        let mut signer = OsslSigner::new(self.scheme.message_digest(), &self.pkey)?;
        if let Some(padding) = self.scheme.padding() {
            signer.set_rsa_padding(padding)?;
        }
        for part in parts {
            signer.update(part)?;
        }
        let mut signature = Signature::empty(self.scheme);
        signer.sign(signature.as_mut_slice())?;
        Ok(signature)
    }
}

impl Verifier for AsymKey<Public, Sign> {
    fn verify<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I, signature: &[u8]) -> Result<()> {
        let mut verifier = OsslVerifier::new(self.scheme.message_digest(), &self.pkey)?;
        if let Some(padding) = self.scheme.padding() {
            verifier.set_rsa_padding(padding)?;
        }
        for part in parts {
            verifier.update(part)?;
        }
        if verifier.verify(signature)? {
            Ok(())
        } else {
            Err(Error::InvalidSignature)
        }
    }
}

#[derive(Clone)]
pub struct AsymKeyPair<S: Scheme> {
    public: AsymKey<Public, S>,
    private: AsymKey<Private, S>,
}

impl<S: Scheme> AsymKeyPair<S> {
    pub fn new(scheme: S, public_der: &[u8], private_der: &[u8]) -> Result<AsymKeyPair<S>> {
        let public = AsymKey::<Public, _>::new(scheme, public_der)?;
        let private = AsymKey::<Private, _>::new(scheme, private_der)?;
        Ok(AsymKeyPair { public, private })
    }
}

// Note that only signing keys are associated with a Principal.
impl Principaled for AsymKeyPair<Sign> {
    fn principal_of_kind(&self, kind: HashKind) -> Principal {
        self.public.principal_of_kind(kind)
    }
}

impl Encrypter for AsymKeyPair<Encrypt> {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        self.public.encrypt(slice)
    }
}

impl Decrypter for AsymKeyPair<Encrypt> {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        self.private.decrypt(slice)
    }
}

impl Signer for AsymKeyPair<Sign> {
    fn sign<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I) -> Result<Signature> {
        self.private.sign(parts)
    }
}

impl Verifier for AsymKeyPair<Sign> {
    fn verify<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I, signature: &[u8]) -> Result<()> {
        self.public.verify(parts, signature)
    }
}

#[derive(Clone)]
pub struct ConcreteCreds {
    sign: AsymKeyPair<Sign>,
    encrypt: AsymKeyPair<Encrypt>,
    writecap: Option<Writecap>,
}

impl ConcreteCreds {
    pub fn new(sign: AsymKeyPair<Sign>, encrypt: AsymKeyPair<Encrypt>) -> ConcreteCreds {
        ConcreteCreds {
            sign,
            encrypt,
            writecap: None,
        }
    }

    #[cfg(test)]
    pub(crate) fn generate() -> Result<ConcreteCreds> {
        let encrypt = Encrypt::RSA_OAEP_3072_SHA_256.generate()?;
        let sign = Sign::RSA_PSS_3072_SHA_256.generate()?;
        Ok(ConcreteCreds {
            sign,
            encrypt,
            writecap: None,
        })
    }
}

impl Verifier for ConcreteCreds {
    fn verify<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I, signature: &[u8]) -> Result<()> {
        self.sign.verify(parts, signature)
    }
}

impl Encrypter for ConcreteCreds {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        self.encrypt.encrypt(slice)
    }
}

impl Principaled for ConcreteCreds {
    fn principal_of_kind(&self, kind: HashKind) -> Principal {
        self.sign.principal_of_kind(kind)
    }
}

impl CredsPub for ConcreteCreds {
    fn public_sign(&self) -> &AsymKey<Public, Sign> {
        &self.sign.public
    }
}

impl Signer for ConcreteCreds {
    fn sign<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I) -> Result<Signature> {
        self.sign.sign(parts)
    }
}

impl Decrypter for ConcreteCreds {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        self.encrypt.decrypt(slice)
    }
}

impl CredsPriv for ConcreteCreds {
    fn writecap(&self) -> Option<&Writecap> {
        self.writecap.as_ref()
    }

    fn set_writecap(&mut self, writecap: Writecap) {
        self.writecap = Some(writecap);
    }
}

impl Creds for ConcreteCreds {}

pub trait Encrypter {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>>;
}

pub trait EncrypterExt: Encrypter {
    fn ser_encrypt<T: Serialize>(&self, value: &T) -> Result<Ciphertext<T>> {
        let data = to_vec(value)?;
        let data = self.encrypt(&data)?;
        Ok(Ciphertext::new(data))
    }
}

impl<T: Encrypter + ?Sized> EncrypterExt for T {}

pub trait Decrypter {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>>;
}

pub trait DecrypterExt: Decrypter {
    fn ser_decrypt<T: DeserializeOwned>(&self, ct: &Ciphertext<T>) -> Result<T> {
        let pt = self.decrypt(ct.data.as_slice())?;
        Ok(from_vec(&pt)?)
    }
}

impl<T: Decrypter + ?Sized> DecrypterExt for T {}

pub trait Signer {
    fn sign<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I) -> Result<Signature>;
}

pub trait SignerExt: Signer {
    fn ser_sign<T: Serialize>(&self, value: &T) -> Result<Signed<T>> {
        let data = to_vec(value)?;
        let sig = self.sign(std::iter::once(data.as_slice()))?;
        Ok(Signed::new(data, sig))
    }

    fn sign_writecap(&self, writecap: &mut Writecap) -> Result<()> {
        let signed = self.ser_sign(&writecap.body)?;
        writecap.signature = signed.sig;
        Ok(())
    }
}

impl<T: Signer + ?Sized> SignerExt for T {}

pub trait Verifier {
    fn verify<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I, signature: &[u8]) -> Result<()>;
}

pub trait VerifierExt: Verifier {
    fn ser_verify<T: Serialize>(&self, value: &T, signature: &[u8]) -> Result<()> {
        let data = to_vec(value)?;
        self.verify(std::iter::once(data.as_slice()), signature)
    }
}

impl<T: Verifier + ?Sized> VerifierExt for T {}

/// Trait for types which can be used as public credentials.
pub trait CredsPub: Verifier + Encrypter + Principaled {
    /// Returns a reference to the public signing key which can be used to verify signatures.
    fn public_sign(&self) -> &AsymKey<Public, Sign>;
}

/// Trait for types which contain private credentials.
pub trait CredsPriv: Decrypter + Signer {
    /// Returns a reference to the writecap associated with these credentials, if one has been
    /// issued.
    fn writecap(&self) -> Option<&Writecap>;

    /// Sets the writecap used by these creds to the given instance.
    fn set_writecap(&mut self, writecap: Writecap);
}

/// Trait for types which contain both public and private credentials.
pub trait Creds: CredsPriv + CredsPub {
    fn issue_writecap(
        &self,
        issued_to: Principal,
        path_components: Vec<String>,
        expires: Epoch,
    ) -> Result<Writecap> {
        let path = BlockPath::new(self.principal(), path_components);
        let body = WritecapBody {
            issued_to,
            path,
            expires,
            signing_key: self.public_sign().to_owned(),
        };
        let signed = self.ser_sign(&body)?;
        Ok(Writecap {
            body,
            signature: signed.sig,
            next: self.writecap().map(|e| Box::new(e.to_owned())),
        })
    }
}

/// A trait for types which store credentials.
pub trait CredStore {
    type CredHandle: Creds;
    type ExportedCreds: Serialize + for<'de> Deserialize<'de>;

    /// Returns the node credentials. If credentials haven't been generated, they are generated
    /// stored and returned.
    fn node_creds(&self) -> Result<Self::CredHandle>;
    /// Returns the root credentials. If no root credentials have been generated, or the provided
    /// password is incorrect, then an error is returned.
    fn root_creds(&self, password: &str) -> Result<Self::CredHandle>;
    /// Generates the root credentials and protects them using the given password. If the root
    /// credentials have already been generated then an error is returned.
    fn gen_root_creds(&self, password: &str) -> Result<Self::CredHandle>;
    fn storage_key(&self) -> Result<AsymKeyPub<Encrypt>>;
    fn export_root_creds(
        &self,
        root_creds: &Self::CredHandle,
        password: &str,
        new_parent: &AsymKeyPub<Encrypt>,
    ) -> Result<Self::ExportedCreds>;
    fn import_root_creds(
        &self,
        password: &str,
        exported: Self::ExportedCreds,
    ) -> Result<Self::CredHandle>;
}

impl BlockMeta {
    /// Validates that this metadata struct contains a valid writecap, that this writecap is
    /// permitted to write to the path of this block and that the signature in this metadata struct
    /// is valid and matches the key the writecap was issued to.
    pub fn assert_valid(&self) -> Result<()> {
        let body = &self.body;
        let writecap = body
            .writecap
            .as_ref()
            .ok_or(crate::Error::MissingWritecap)?;
        writecap.assert_valid_for(&body.path)?;
        let signed_by = body.signing_key.principal();
        if writecap.body.issued_to != signed_by {
            return Err(Error::SignatureMismatch {
                actual: signed_by,
                expected: writecap.body.issued_to.clone(),
            });
        }
        body.signing_key.ser_verify(&body, self.sig.as_slice())
    }
}

/// The types of errors which can occur when verifying a writecap chain is authorized to write to
/// a given path.
#[derive(Debug, PartialEq, Display)]
pub enum WritecapAuthzErr {
    /// The chain is not valid for use on the given path.
    UnauthorizedPath,
    /// At least one writecap in the chain is expired.
    Expired,
    /// The given writecaps do not actually form a chain.
    NotChained,
    /// The principal the root writecap was issued to does not own the given path.
    RootDoesNotOwnPath,
    /// An error occurred while serializing a writecap.
    Serde(String),
    /// The write cap chain was too long to be validated. The value contained in this error is
    /// the maximum allowed length.
    ChainTooLong(usize),
}

impl Writecap {
    /// Verifies that the given `Writecap` actually grants permission to write to the given `Path`.
    pub fn assert_valid_for(&self, path: &BlockPath) -> Result<()> {
        let mut writecap = self;
        const CHAIN_LEN_LIMIT: usize = 256;
        let mut prev: Option<&Writecap> = None;
        let mut sig_input_buf = Vec::new();
        let now = Epoch::now();
        for _ in 0..CHAIN_LEN_LIMIT {
            if !writecap.body.path.contains(path) {
                return Err(WritecapAuthzErr::UnauthorizedPath.into());
            }
            if writecap.body.expires <= now {
                return Err(WritecapAuthzErr::Expired.into());
            }
            if let Some(prev) = &prev {
                if prev
                    .body
                    .signing_key
                    .principal_of_kind(writecap.body.issued_to.kind())
                    != writecap.body.issued_to
                {
                    return Err(WritecapAuthzErr::NotChained.into());
                }
            }
            sig_input_buf.clear();
            write_to(&writecap.body, &mut sig_input_buf)
                .map_err(|e| WritecapAuthzErr::Serde(e.to_string()))?;
            writecap.body.signing_key.verify(
                std::iter::once(sig_input_buf.as_slice()),
                writecap.signature.as_slice(),
            )?;
            match &writecap.next {
                Some(next) => {
                    prev = Some(writecap);
                    writecap = next;
                }
                None => {
                    // We're at the root key. As long as the signer of this writecap is the owner of
                    // the path, then the writecap is valid.
                    if writecap
                        .body
                        .signing_key
                        .principal_of_kind(path.root().kind())
                        == *path.root()
                    {
                        return Ok(());
                    } else {
                        return Err(WritecapAuthzErr::RootDoesNotOwnPath.into());
                    }
                }
            }
        }
        Err(WritecapAuthzErr::ChainTooLong(CHAIN_LEN_LIMIT).into())
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::{
        crypto::secret_stream::SecretStream, test_helpers::*, BrotliParams, Sectored, SectoredBuf,
        TryCompose,
    };
    use std::{
        io::{Seek, SeekFrom},
        time::Duration,
    };

    #[test]
    fn encrypt_decrypt_block() {
        const SECT_SZ: usize = 16;
        const SECT_CT: usize = 8;
        let key = make_key_pair();
        let readcap = make_readcap_for(&key);
        let mut block = make_block_with(readcap);
        write_fill(&mut block, SECT_SZ, SECT_CT);
        block.seek(SeekFrom::Start(0)).expect("seek failed");
        read_check(block, SECT_SZ, SECT_CT);
    }

    #[test]
    fn rsa_sign_and_verify() -> Result<()> {
        let key = make_key_pair();
        let header = b"About: lyrics".as_slice();
        let message = b"Everything that feels so good is bad bad bad.".as_slice();
        let signature = key.sign([header, message].into_iter())?;
        key.verify([header, message].into_iter(), signature.as_slice())
    }

    #[test]
    fn hash_to_string() {
        let hash = make_principal().0;
        let string = hash.to_string();
        assert_eq!(
            "Sha2_256:dSip4J0kurN5VhVo_aTipM-ywOOWrqJuRRVQ7aa-bew",
            string
        )
    }

    #[test]
    fn hash_to_string_round_trip() -> Result<()> {
        let expected = make_principal().0;
        let string = expected.to_string();
        let actual = Hash::try_from(string.as_str())?;
        assert_eq!(expected, actual);
        Ok(())
    }

    #[test]
    fn verify_writecap_valid() {
        let writecap = make_writecap(vec!["apps", "verse"]);
        writecap
            .assert_valid_for(&writecap.body.path)
            .expect("failed to verify writecap");
    }

    #[test]
    fn verify_writecap_invalid_signature() -> Result<()> {
        let mut writecap = make_writecap(vec!["apps", "verse"]);
        writecap.signature = Signature::empty(Sign::RSA_PSS_3072_SHA_256);
        let result = writecap.assert_valid_for(&writecap.body.path);
        if let Err(Error::InvalidSignature) = result {
            Ok(())
        } else {
            Err(Error::custom(format!("unexpected result {:?}", result)))
        }
    }

    fn assert_authz_err<T: std::fmt::Debug>(
        expected: WritecapAuthzErr,
        result: Result<T>,
    ) -> Result<()> {
        if let Err(Error::WritecapAuthzErr(actual)) = &result {
            if *actual == expected {
                return Ok(());
            }
        }
        Err(Error::custom(format!("unexpected result: {:?}", result)))
    }

    #[test]
    fn verify_writecap_invalid_path_not_contained() -> Result<()> {
        let writecap = make_writecap(vec!["apps", "verse"]);
        let mut path = writecap.body.path.clone();
        path.pop_component();
        // `path` is now a superpath of `writecap.path`, thus the writecap is not authorized to
        // write to it.
        let result = writecap.assert_valid_for(&path);
        assert_authz_err(WritecapAuthzErr::UnauthorizedPath, result)
    }

    #[test]
    fn verify_writecap_invalid_expired() -> Result<()> {
        let mut writecap = make_writecap(vec!["apps", "verse"]);
        writecap.body.expires = Epoch::now() - Duration::from_secs(1);
        let result = writecap.assert_valid_for(&writecap.body.path);
        assert_authz_err(WritecapAuthzErr::Expired, result)
    }

    #[test]
    fn verify_writecap_invalid_not_chained() -> Result<()> {
        let (mut root_writecap, root_key) = make_self_signed_writecap();
        root_writecap.body.issued_to = Principal(Hash::Sha2_256([0; 32]));
        root_key.sign_writecap(&mut root_writecap)?;
        let node_principal = NODE_CREDS.principal();
        let writecap = make_writecap_trusted_by(
            root_writecap,
            &root_key,
            node_principal,
            vec!["apps", "contacts"],
        );
        let result = writecap.assert_valid_for(&writecap.body.path);
        assert_authz_err(WritecapAuthzErr::NotChained, result)
    }

    #[test]
    fn verify_writecap_invalid_root_doesnt_own_path() -> Result<()> {
        let (mut root_writecap, root_key) = make_self_signed_writecap();
        let owner = Principal(Hash::Sha2_256([0; 32]));
        root_writecap.body.path = make_path_with_root(owner, vec![]);
        root_key.sign_writecap(&mut root_writecap)?;
        let node_principal = NODE_CREDS.principal();
        let writecap = make_writecap_trusted_by(
            root_writecap,
            &root_key,
            node_principal,
            vec!["apps", "contacts"],
        );
        let result = writecap.assert_valid_for(&writecap.body.path);
        assert_authz_err(WritecapAuthzErr::RootDoesNotOwnPath, result)
    }

    #[test]
    fn aeadkey_encrypt_decrypt_aes256gcm() {
        let key = AeadKey::new(AeadKeyKind::AesGcm256).expect("failed to create key");
        let aad = [1u8; 16];
        let expected = [2u8; 32];
        let tagged = key.encrypt(aad, &expected).expect("encrypt failed");
        let actual = key.decrypt(&tagged).expect("decrypt failed");
        assert_eq!(expected, actual.as_slice());
    }

    #[test]
    fn aeadkey_decrypt_fails_when_ct_modified() {
        let key = AeadKey::new(AeadKeyKind::AesGcm256).expect("failed to create key");
        let aad = [1u8; 16];
        let expected = [2u8; 32];
        let mut tagged = key.encrypt(aad, &expected).expect("encrypt failed");
        tagged.ciphertext.data[0] = tagged.ciphertext.data[0].wrapping_add(1);
        let result = key.decrypt(&tagged);
        assert!(result.is_err())
    }

    #[test]
    fn aeadkey_decrypt_fails_when_aad_modified() {
        let key = AeadKey::new(AeadKeyKind::AesGcm256).expect("failed to create key");
        let aad = [1u8; 16];
        let expected = [2u8; 32];
        let mut tagged = key.encrypt(aad, &expected).expect("encrypt failed");
        tagged.aad[0] = tagged.aad[0].wrapping_add(1);
        let result = key.decrypt(&tagged);
        assert!(result.is_err())
    }

    #[test]
    fn compose_merkle_and_secret_streams() {
        use merkle_stream::tests::make_merkle_stream_filled_with_zeros;
        const SECT_SZ: usize = 4096;
        const SECT_CT: usize = 16;
        let merkle = make_merkle_stream_filled_with_zeros(SECT_SZ, SECT_CT);
        let key = SymKey::generate(SymKeyKind::Aes256Cbc).expect("key generation failed");
        let mut secret = SecretStream::new(key)
            .try_compose(merkle)
            .expect("compose for secret failed");
        let secret_sect_sz = secret.sector_sz();
        write_fill(&mut secret, secret_sect_sz, SECT_CT);
        secret.seek(SeekFrom::Start(0)).expect("seek failed");
        read_check(&mut secret, secret_sect_sz, SECT_CT);
    }

    #[test]
    fn compress_then_encrypt() {
        use brotli::{CompressorWriter, Decompressor};
        const SECT_SZ: usize = 4096;
        const SECT_CT: usize = 16;
        let params = BrotliParams::new(SECT_SZ, 8, 20);
        let key = SymKey::generate(SymKeyKind::Aes256Cbc).expect("key generation failed");
        let mut inner = SectoredBuf::new()
            .try_compose(
                SecretStream::new(key)
                    .try_compose(SectoredCursor::new(Vec::new(), SECT_SZ))
                    .expect("compose for inner failed"),
            )
            .expect("compose with sectored buffer failed");
        {
            let write: CompressorWriter<_> = params
                .clone()
                .try_compose(&mut inner)
                .expect("compose for write failed");
            write_fill(write, SECT_SZ, SECT_CT);
        }
        inner.seek(SeekFrom::Start(0)).expect("seek failed");
        {
            let read: Decompressor<_> = params
                .try_compose(&mut inner)
                .expect("compose for read failed");
            read_check(read, SECT_SZ, SECT_CT);
        }
    }

    /// Tests that validate the dependencies of this module.
    mod dependency_tests {
        use super::*;
        use openssl::{
            ec::{EcGroup, EcKey},
            nid::Nid,
        };

        /// This test validates that data encrypted with AES 256 CBC can later be decrypted.
        #[test]
        fn aes_256_cbc_roundtrip() {
            use super::*;
            let expected = b"We attack at the crack of noon!";
            let cipher = Cipher::aes_256_cbc();

            let key = BLOCK_KEY.key_slice();
            let iv = BLOCK_KEY.iv_slice();
            let ciphertext = openssl_encrypt(cipher, key, iv, expected).unwrap();
            let actual = openssl_decrypt(cipher, key, iv, ciphertext.as_slice()).unwrap();

            assert_eq!(expected, actual.as_slice());
        }

        /// Tests that the keys for the SECP256K1 curve are the expected sizes.
        #[test]
        fn secp256k1_key_lengths() {
            let group = EcGroup::from_curve_name(Nid::SECP256K1).unwrap();
            let key = EcKey::generate(&group).unwrap();
            let public = key.public_key_to_der().unwrap();
            let private = key.private_key_to_der().unwrap();
            let public_len = public.len();
            let private_len = private.len();
            assert_eq!(88, public_len);
            assert_eq!(118, private_len);
        }

        #[test]
        fn ed25519_key_lengths() {
            let key = PKey::generate_x25519().unwrap();
            let public = key.public_key_to_der().unwrap();
            let private = key.private_key_to_der().unwrap();
            let public_len = public.len();
            let private_len = private.len();
            assert_eq!(44, public_len);
            assert_eq!(48, private_len);
        }
    }
}
